package gov.usgs.earthquake.nshmp.netcdf;

import static com.google.common.base.Preconditions.checkArgument;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.logging.Level;
import java.util.logging.Logger;

import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.gmm.Imt;
import gov.usgs.earthquake.nshmp.netcdf.reader.BoundingHazards;
import gov.usgs.earthquake.nshmp.netcdf.reader.BoundingHazardsReader;
import gov.usgs.earthquake.nshmp.netcdf.reader.NetcdfCoordinates;
import gov.usgs.earthquake.nshmp.netcdf.reader.StaticHazard;
import gov.usgs.earthquake.nshmp.netcdf.reader.StaticHazards;

import ucar.nc2.dataset.NetcdfDataset;

/**
 *
 * Hazard curves stored in netCDF
 *
 * This class is currently specific to NSHMP 2018a data release.
 *
 * TODO: generalize for any NSHMP model data release with subclasses per data
 * release.
 *
 * TODO: Replace loops with streams
 *
 * TODO: Utilize XySequence.copyOf()
 *
 * TODO: Update netCDF file format, rename variables, implement enums,
 * standardize attributes, remove Vs30 dimension from Imls
 *
 * TODO: attempt to calculate index, possibly as
 * NshmpNetcdfCoordinates.calcIdxLTEQ()
 *
 * @author U.S. Geological Survey
 */
public class NshmNetcdfReader {

  private static final Logger LOGGER = Logger.getLogger("ucar");

  private final Path path;
  private final NshmGroup nshmGroup;
  private final NetcdfCoordinates coords;

  static {
    /* Update ucar logger */
    LOGGER.setLevel(Level.SEVERE);
  }

  public NshmNetcdfReader(NshmGroup nshmGroup, Path path) {
    this.nshmGroup = nshmGroup;
    this.path = path;

    if (Files.notExists(path)) {
      throw new IllegalArgumentException("Path to Netcdf file [" + path + "] does not exist");
    }

    try (var ncd = NetcdfDataset.openDataset(path.toString())) {
      // TODO:
      // - Handle metadata from netCDF file get root group and
      // attributes, and attributes of subgroups
      // - Get netCDF dimensions
      var targetGroup = ncd.findGroup(nshmGroup.baseGroup());
      coords = new NetcdfCoordinates(nshmGroup, targetGroup);
    } catch (IOException e) {
      throw new RuntimeException("Could not read Netcdf file [" + path + " ]");
    }
  }

  /**
   * Returns the NSHM NetCDF group
   */
  public NshmGroup nshmGroup() {
    return nshmGroup;
  }

  /**
   * Returns the NetCDF dimensions and coordinate variables.
   */
  public NetcdfCoordinates coordinates() {
    return coords;
  }

  /**
   * Returns the path to the NetCDF file.
   */
  public Path path() {
    return path;
  }

  /**
   * Return the full set of hazard curves at the bounding grid points and the
   * target Location. Intended for validation uses.
   *
   * @param site The location to get bounding hazards
   */
  public BoundingHazards boundingHazards(Location site) {
    return BoundingHazardsReader.boundingHazards(this, site);
  }

  /**
   * Return a {@code Map<SiteClass, Map<Imt, XySequence>>} with hazard curves at
   * the specified Location.
   *
   * @param site The site to get the hazard curves
   */
  public StaticHazards hazard(Location site) {
    return boundingHazards(site).get(site);
  }

  /**
   * Return a hazard curve for each Imt at a specific site class.
   *
   * @param site The site to get hazard curves
   * @param siteClass The site class
   */
  public StaticHazard hazard(Location site, SiteClass siteClass) {
    checkArgument(
        coords.siteClasses().contains(siteClass),
        "Site class [" + siteClass + "] not supported");
    return hazard(site).get(siteClass);
  }

  /**
   * Return a single hazard curve for the specified Location, SiteClass, and
   * Imt.
   *
   * @param site The site to get the hazard curves
   * @param siteClass The site class
   * @param imt The IMT
   */
  public XySequence hazard(Location site, SiteClass siteClass, Imt imt) {
    checkArgument(coords.imts().contains(imt), "Imt [" + imt + "] not supported");
    return hazard(site, siteClass).get(imt);
  }

}
