package gov.usgs.earthquake.nshmp.netcdf.www;

import static gov.usgs.earthquake.nshmp.netcdf.www.NetcdfWsUtils.GSON;

import java.nio.file.Path;
import java.util.List;
import java.util.logging.Logger;
import java.util.stream.Collectors;

import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.geo.Bounds;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.json.Feature;
import gov.usgs.earthquake.nshmp.geo.json.GeoJson;
import gov.usgs.earthquake.nshmp.gmm.Imt;
import gov.usgs.earthquake.nshmp.internal.www.NshmpMicronautServlet.UrlHelper;
import gov.usgs.earthquake.nshmp.internal.www.Response;
import gov.usgs.earthquake.nshmp.internal.www.WsUtils;
import gov.usgs.earthquake.nshmp.internal.www.meta.Status;
import gov.usgs.earthquake.nshmp.netcdf.NshmGroup;
import gov.usgs.earthquake.nshmp.netcdf.NshmNetcdfReader;
import gov.usgs.earthquake.nshmp.netcdf.SiteClass;
import gov.usgs.earthquake.nshmp.netcdf.reader.BoundingHazards;
import gov.usgs.earthquake.nshmp.netcdf.reader.StaticHazard;
import gov.usgs.earthquake.nshmp.netcdf.reader.StaticHazards;
import gov.usgs.earthquake.nshmp.netcdf.www.NetcdfController.Query;
import gov.usgs.earthquake.nshmp.netcdf.www.NetcdfController.Service;
import gov.usgs.earthquake.nshmp.netcdf.www.NetcdfWsUtils.Key;

import io.micronaut.http.HttpResponse;

/**
 * Service handler for {@code NetcdfController}.
 * 
 * @see NetcdfController
 * 
 * @author U.S. Geological Survey
 */
public class NetcdfService {

  private static NshmNetcdfReader netcdf;

  private static final String SERVICE_NAME = "Netcdf NSHM Service";
  private static final String SERVICE_DESCRIPTION = "Get static hazard curves from a NetCDF file";
  private static final String SERVICE_SYNTAX = "%s/";

  private static final Logger LOGGER = Logger.getLogger(NetcdfService.class.getName());

  static HttpResponse<String> handleDoGet(
      NshmGroup nshmGroup,
      Path path,
      UrlHelper urlHelper,
      Query query) {
    try {
      setNetcdf(nshmGroup, path);
      LOGGER.info(SERVICE_NAME + " - Request: " + urlHelper.url);
      LOGGER.info(SERVICE_NAME + " - Query:\n" + GSON.toJson(query));

      if (query.longitude == null && query.latitude == null) {
        return metadata(nshmGroup, path, urlHelper);
      }
      var service = getService(nshmGroup, path, query, urlHelper);
      Response<?, ?> svcResponse = processRequest(netcdf, service, query, urlHelper);
      var response = GSON.toJson(svcResponse);
      LOGGER.info(SERVICE_NAME + " - Result:\n" + response);
      return HttpResponse.ok(response);
    } catch (Exception e) {
      return NetcdfWsUtils.handleError(e, SERVICE_NAME, LOGGER, urlHelper);
    }
  }

  static HttpResponse<String> handleDoGetBounding(
      NshmGroup nshmGroup,
      Path path,
      UrlHelper urlHelper,
      Query query) {
    try {
      setNetcdf(nshmGroup, path);
      LOGGER.info(SERVICE_NAME + " - Request: " + urlHelper.url);
      LOGGER.info(SERVICE_NAME + " - Query:\n" + GSON.toJson(query));

      if (query.longitude == null && query.latitude == null) {
        return metadata(nshmGroup, path, urlHelper);
      }
      Response<?, ?> svcResponse = processRequest(netcdf, Service.BOUNDING, query, urlHelper);
      var response = GSON.toJson(svcResponse);
      LOGGER.info(SERVICE_NAME + " - Result:\n" + response);
      return HttpResponse.ok(response);
    } catch (Exception e) {
      return NetcdfWsUtils.handleError(e, SERVICE_NAME, LOGGER, urlHelper);
    }

  }

  static HttpResponse<String> handleDoGetBorder(
      NshmGroup nshmGroup,
      Path path,
      UrlHelper urlHelper) {
    try {
      setNetcdf(nshmGroup, path);
      var geojson = GeoJson.builder();
      var border = netcdf.coordinates().region().border();

      var id = 0;
      for (var location : border) {
        var feature = Feature.point(location)
            .id(id++)
            .build();

        geojson.add(feature);
      }

      return HttpResponse.ok(geojson.toJson());
    } catch (Exception e) {
      return NetcdfWsUtils.handleError(e, SERVICE_NAME, LOGGER, urlHelper);
    }
  }

  static HttpResponse<String> metadata(NshmGroup nshmGroup, Path path, UrlHelper urlHelper) {
    try {
      setNetcdf(nshmGroup, path);
      String netcdfFile = path == null ? "" : path.toString();
      var metadata = new Metadata(netcdfFile, urlHelper);
      var svcResponse =
          new Response<>(Status.USAGE, SERVICE_NAME, urlHelper.url, metadata, urlHelper);
      var response = GSON.toJson(svcResponse);
      LOGGER.info(SERVICE_NAME + " - Result:\n" + response);
      return HttpResponse.ok(response);
    } catch (Exception e) {
      return NetcdfWsUtils.handleError(e, SERVICE_NAME, LOGGER, urlHelper);
    }
  }

  static Service getService(NshmGroup nshmGroup, Path path, Query query, UrlHelper urlHelper) {
    if (query.siteClass != null && query.imt != null) {
      return Service.CURVES_BY_SITE_CLASS_IMT;
    } else if (query.siteClass != null && query.imt == null) {
      return Service.CURVES_BY_SITE_CLASS;
    } else {
      return Service.CURVES;
    }
  }

  static Response<?, ?> processRequest(
      NshmNetcdfReader netcdf,
      Service service,
      Query query,
      UrlHelper urlHelper) {
    RequestData requestData = new RequestData(query.longitude, query.latitude);

    switch (service) {
      case BOUNDING:
        return processBounding(requestData, urlHelper);
      case CURVES:
        return processCurves(requestData, urlHelper);
      case CURVES_BY_SITE_CLASS:
        requestData = new RequestDataCurves(query.longitude, query.latitude, query.siteClass);
        return processCurvesSiteClass(
            (RequestDataCurves) requestData,
            urlHelper);
      case CURVES_BY_SITE_CLASS_IMT:
        requestData = new RequestDataCurve(
            query.longitude, query.latitude, query.siteClass, query.imt);
        return processCurve((RequestDataCurve) requestData, urlHelper);
      default:
        throw new RuntimeException("Netcdf service [" + service + "] not found");
    }
  }

  static Response<RequestData, List<List<List<ResponseData>>>> processBounding(
      RequestData request,
      UrlHelper urlHelper) {
    WsUtils.checkValue(Key.LATITUDE, request.site.latitude);
    WsUtils.checkValue(Key.LONGITUDE, request.site.longitude);
    var bounding = netcdf.boundingHazards(request.site);
    var boundingAsList = toLists(netcdf.nshmGroup(), request.site, bounding);
    return new Response<>(Status.SUCCESS, SERVICE_NAME, request, boundingAsList, urlHelper);
  }

  static Response<RequestData, List<List<ResponseData>>> processCurves(
      RequestData request,
      UrlHelper urlHelper) {
    WsUtils.checkValue(Key.LATITUDE, request.site.latitude);
    WsUtils.checkValue(Key.LONGITUDE, request.site.longitude);
    var curves = netcdf.hazard(request.site);
    var curvesAsList = toList(netcdf.nshmGroup(), request.site, curves);
    return new Response<>(Status.SUCCESS, SERVICE_NAME, request, curvesAsList, urlHelper);
  }

  static Response<RequestDataCurves, List<ResponseData>> processCurvesSiteClass(
      RequestDataCurves request,
      UrlHelper urlHelper) {
    WsUtils.checkValue(Key.LATITUDE, request.site.latitude);
    WsUtils.checkValue(Key.LONGITUDE, request.site.longitude);
    WsUtils.checkValue(Key.SITE_CLASS, request.siteClass);
    var curves = netcdf.hazard(request.site, request.siteClass);
    var curvesAsList = toList(netcdf.nshmGroup(), request, curves);
    return new Response<>(Status.SUCCESS, SERVICE_NAME, request, curvesAsList, urlHelper);
  }

  static Response<RequestDataCurve, ResponseData> processCurve(
      RequestDataCurve request,
      UrlHelper urlHelper) {
    WsUtils.checkValue(Key.LATITUDE, request.site.latitude);
    WsUtils.checkValue(Key.LONGITUDE, request.site.longitude);
    WsUtils.checkValue(Key.SITE_CLASS, request.siteClass);
    WsUtils.checkValue(Key.IMT, request.imt);

    var site = request.site;
    var metadata = new ResponseMetadata(
        netcdf.nshmGroup(), site.longitude, site.latitude, request.siteClass, request.imt);
    var curve = netcdf.hazard(request.site, request.siteClass, request.imt);
    var responseData = new ResponseData(metadata, curve);
    return new Response<>(Status.SUCCESS, SERVICE_NAME, request, responseData, urlHelper);
  }

  static List<List<List<ResponseData>>> toLists(
      NshmGroup nshmGroup,
      Location site,
      BoundingHazards bounding) {

    return bounding.entrySet().stream()
        .map(entry -> toList(nshmGroup, site, entry.getValue()))
        .collect(Collectors.toList());
  }

  static List<List<ResponseData>> toList(
      NshmGroup nshmGroup,
      Location site,
      StaticHazards curves) {
    return curves.entrySet().stream()
        .map(entry -> {
          var request = new RequestDataCurves(site.longitude, site.latitude, entry.getKey());
          return toList(nshmGroup, request, entry.getValue());
        })
        .collect(Collectors.toList());
  }

  static List<ResponseData> toList(
      NshmGroup nshmGroup,
      RequestDataCurves request,
      StaticHazard curves) {
    return curves.entrySet().stream()
        .map((entry) -> {
          var site = request.site;
          var metadata = new ResponseMetadata(
              nshmGroup, site.longitude, site.latitude, request.siteClass, entry.getKey());
          return new ResponseData(metadata, entry.getValue());
        })
        .collect(Collectors.toList());
  }

  static class RequestData {
    Location site;

    RequestData(double longitude, double latitude) {
      site = Location.create(latitude, longitude);
    }
  }

  static class RequestDataCurves extends RequestData {
    SiteClass siteClass;

    RequestDataCurves(double longitude, double latitude, SiteClass siteClass) {
      super(longitude, latitude);
      this.siteClass = siteClass;
    }
  }

  static class RequestDataCurve extends RequestDataCurves {
    Imt imt;

    RequestDataCurve(double longitude, double latitude, SiteClass siteClass, Imt imt) {
      super(longitude, latitude, siteClass);
      this.imt = imt;
    }
  }

  static class ResponseMetadata extends RequestDataCurve {
    final String xLabel;
    final String yLabel;
    final NshmGroup nshm;

    ResponseMetadata(
        NshmGroup nshmGroup,
        double longitude,
        double latitude,
        SiteClass siteClass,
        Imt imt) {
      super(longitude, latitude, siteClass, imt);
      nshm = nshmGroup;
      xLabel = "Ground Motion (g)";
      yLabel = "Annual Frequency of Exceedence";
    }
  }

  static class ResponseData {
    final ResponseMetadata metadata;
    final XySequence data;

    ResponseData(ResponseMetadata metadata, XySequence data) {
      this.metadata = metadata;
      this.data = data;
    }
  }

  static class Metadata {

    final String description;
    final String syntax;
    final Parameters parameters;

    Metadata(String netcdfFile, UrlHelper urlHelper) {
      description = SERVICE_DESCRIPTION;
      syntax = String.format(SERVICE_SYNTAX, urlHelper.urlPrefix);
      parameters = new Parameters(netcdfFile);
    }
  }

  static class Parameters {
    final String netcdfFile;
    final NshmGroup nshm;
    final Bounds bounds;
    final List<SiteClass> siteClasses;
    final List<Imt> imts;

    Parameters(String netcdfFile) {
      this.netcdfFile = netcdfFile;
      nshm = netcdf.nshmGroup();
      var coordinates = netcdf.coordinates();
      bounds = coordinates.region().bounds();
      siteClasses = coordinates.siteClasses();
      imts = coordinates.imts();
    }

  }

  private static void setNetcdf(NshmGroup nshmGroup, Path path) {
    if (netcdf == null) {
      netcdf = new NshmNetcdfReader(nshmGroup, path);
    }
  }

}
